# -*- coding: utf-8 -*-
"""define some const terms
Author  : NextPCG
"""

dson_inputs_tag = "Inputs"
dson_input_tag = "input"
dson_params_tag = "Params"
dson_outputs_tag = "Outputs"
dson_output_tag = "output"
dson_doc_tag = "Doc"
dson_meta_field_tag = "DsonMeta"
dson_meta_clsname_tag = "clsName"
dson_meta_if_in_server_tag = "ifInServer"
dson_meta_plugin_name = "pluginName"



# -*- coding: utf-8 -*-
"""
Author  : NextPCG
"""

import json
import os
import time
from typing import Dict, List, Tuple, Any
from inspect import signature, Signature, Parameter
import logging
from abc import ABCMeta, abstractmethod
from enum import Enum
from .meta_helper import Singleton

"""note:
io_mode:
0: server_text (like houdini_text)
1: houdini_text (support)
2: houdini_binary (support)
3: houdini_hdk (not support)
 """


class FieldDesc:
    """
    helper struct used when create field or get outputs from field
    """

    def __init__(self, work_path, io_mode=1, raw_compress=False) -> None:
        self.work_path: str = work_path
        self.raw_compress: bool = raw_compress
        self.io_mode: int = io_mode


def unique_name():
    n = 1
    while True:
        yield n
        n += 1
        if n > 10000:
            n = 1


unique_name_generator = unique_name()

'''
FieldManager   
'''


class FieldManager(metaclass=Singleton):
    def __init__(self):
        self.fieldMap = {}  # map from field_type_name(str) to actual Field

    def create_field(self, json_object, input_name, field_desc: FieldDesc):
        """create a Field

        Args:
            json_object(Dict)
            work_path(str)
        Return:
            Field
            """
        # try:
        assert "If List" in json_object
        if json_object['If List']:
            return ListField.from_json(json_object, input_name, field_desc)
        else:
            return self.fieldMap[json_object['Type']].from_json(json_object, input_name, field_desc)
        # except Exception as e:
        #     logging.exception("get_input failed, json_object:" + str(json_object))

    def __getitem__(self, item):
        return self.fieldMap[item]


'''
define Field
'''

BOOL_TYPE = 'paramType.Bool'
INT_TYPE = 'paramType.Int'
INT2_TYPE = 'paramType.Int2'
INT3_TYPE = 'paramType.Int3'
INT4_TYPE = 'paramType.Int4'
FLOAT_TYPE = 'paramType.Float'
FLOAT2_TYPE = 'paramType.Float2'
FLOAT3_TYPE = 'paramType.Float3'
FLOAT4_TYPE = 'paramType.Float4'
STRING_TYPE = 'paramType.String'
JSON_TYPE = 'paramType.Json'
FILE_TYPE = 'paramType.File'
LIST_TYPE = 'paramType.List'


class FieldCategory(Enum):
    Input = 0
    Param = 1
    Output = 2


class FieldMeta(type):
    def __init__(cls, *args, **kwargs):
        assert hasattr(cls, 'field_type_name')
        FieldManager().fieldMap[cls.field_type_name] = cls
        super().__init__(*args, **kwargs)


"""some notes about Field
1. field_type_name will be registered in FieldManager. So FiledManager Can create Field according to 'Type' field in json object
2. create_json_field(): for supporting generate json from function's signature
3. from_json : for creating field from json field in Inputs or Params
4. get_input: functions use this method to get actual value to impl_function
5. get_output: fill the 'Value Data' hole in json_field
"""


class Field(metaclass=FieldMeta):
    field_type_name = ""
    if_list = False
    support_param = True  # if this field can be a param(with default value)

    @classmethod
    def create_json_field(cls, field_category=FieldCategory.Input, name: str = None):
        """ create a json object
        """
        """get json format field"""
        field = {'Type': cls.field_type_name, 'If List': cls.if_list, 'FieldCategory': field_category.name}
        if name is not None:
            field["Label"] = name
        return field

    def __init__(self, value=None):
        self._value = value

    @classmethod
    def from_json(cls, json_object, input_name, field_desc: FieldDesc):
        """
        Args:
            json_object: the json field describe this field type in json(dson) file
        """
        return cls(json_object['Value Data'])

    def get_input(self):
        """ pass input to impl function"""
        self.type_assert()
        return self._value

    def type_assert(self):
        pass

    def get_output(self, json_field, field_desc: FieldDesc, data_field='Value Data', output_index=0):
        json_field[data_field] = self._value

    @classmethod
    def get_param_default(cls, json_object, default_value):
        """fill default value to json field"""
        json_object["Default Value"] = default_value


# base class for those who want to support list field
class ListSubFieldBase:
    @classmethod
    def list_from_json_value(cls, json_value, work_path=None):
        return json_value

    @classmethod
    def list_to_json_value(cls, val, work_path=None):
        return val


class Bool(Field):
    field_type_name = BOOL_TYPE

    @classmethod
    def from_json(cls, json_object, input_name, field_desc: FieldDesc):
        return cls(bool(json_object['Value Data'][0]))

    def get_output(self, json_field, field_desc: FieldDesc, data_field='Value Data', output_index=0):
        # json_field[data_field] = [str(self._value)]
        json_field[data_field] = [self._value]


class Int(Field, ListSubFieldBase):
    field_type_name = INT_TYPE

    def type_assert(self):
        assert isinstance(self._value, int)

    @classmethod
    def from_json(cls, json_object, input_name, field_desc: FieldDesc):
        return cls(int(json_object['Value Data'][0]))

    def get_output(self, json_field, field_desc: FieldDesc, data_field='Value Data', output_index=0):
        # json_field[data_field] = [str(self._value)]
        json_field[data_field] = [self._value]

    @classmethod
    def get_param_default(cls, json_object, default_value):
        json_object["Default Value"] = [default_value]


class Int2(Field, ListSubFieldBase):
    field_type_name = INT2_TYPE

    def type_assert(self):
        assert isinstance(self._value, tuple) or isinstance(self._value, list)
        if len(self._value) != 2:
            raise Exception("value is not int2: " + str(self._value))

    @classmethod
    def from_json(cls, json_object, input_name, field_desc: FieldDesc):
        return cls(list(map(lambda a: int(a), json_object['Value Data'])))

    def get_output(self, json_field, field_desc: FieldDesc, data_field='Value Data', output_index=0):
        # json_field[data_field] = list(map(lambda a: str(a), self._value))
        json_field[data_field] = self._value


class Int3(Int2):
    field_type_name = INT3_TYPE

    def type_assert(self):
        assert isinstance(self._value, tuple) or isinstance(self._value, list)
        if len(self._value) != 3:
            raise Exception("value is not int3: " + str(self._value))


class Int4(Int2):
    field_type_name = INT4_TYPE

    def type_assert(self):
        assert isinstance(self._value, tuple) or isinstance(self._value, list)
        if len(self._value) != 4:
            raise Exception("value is not int4: " + str(self._value))


class Float(Field):
    field_type_name = FLOAT_TYPE

    def type_assert(self):
        assert isinstance(self._value, float)

    @classmethod
    def from_json(cls, json_object, input_name, field_desc: FieldDesc):
        return cls(float(json_object['Value Data'][0]))

    def get_output(self, json_field, field_desc: FieldDesc, data_field='Value Data', output_index=0):
        # json_field[data_field] = [str(self._value)]
        json_field[data_field] = [self._value]

    @classmethod
    def get_param_default(cls, json_object, default_value):
        assert isinstance(default_value, float)
        json_object['Default Value'] = [default_value]


class Float2(Field, ListSubFieldBase):
    field_type_name = FLOAT2_TYPE

    def type_assert(self):
        assert isinstance(self._value, tuple) or isinstance(self._value, list)
        if len(self._value) != 2:
            raise Exception("value is not float2: " + str(self._value))

    @classmethod
    def from_json(cls, json_object, input_name, field_desc: FieldDesc):
        # return cls(list(map(lambda a: float(a), json_object['Value Data'])))
        return cls(json_object['Value Data'])

    def get_output(self, json_field, field_desc: FieldDesc, data_field='Value Data', output_index=0):
        # json_field['Value Data'] = list(map(lambda a: str(a), self._value))
        json_field[data_field] = self._value


class Float3(Float2, ListSubFieldBase):
    field_type_name = FLOAT3_TYPE

    def type_assert(self):
        assert isinstance(self._value, tuple) or isinstance(self._value, list)
        if len(self._value) != 3:
            raise Exception("value is not float2:" + str(self._value))


class Float4(Float2):
    field_type_name = FLOAT4_TYPE

    def type_assert(self):
        assert isinstance(self._value, tuple) or isinstance(self._value, list)
        if len(self._value) != 4:
            raise Exception("value is not float2:" + str(self._value))


class String(Field, ListSubFieldBase):
    field_type_name = STRING_TYPE

    def type_assert(self):
        assert isinstance(self._value, str)

    @classmethod
    def create_json_field(cls, field_category=FieldCategory.Input, name: str = None):
        field = super().create_json_field(field_category, name)
        field["String Type"] = "stringParmType.Regular"
        return field

    @classmethod
    def from_json(cls, json_object, input_name, field_desc: FieldDesc):
        return cls(json_object['Value Data'][0])

    def get_output(self, json_field, field_desc: FieldDesc, data_field='Value Data', output_index=0):
        json_field[data_field] = [self._value]

    @classmethod
    def get_param_default(cls, json_object, default_value):
        assert isinstance(default_value, str)
        json_object['Default Value'] = [default_value]


# _value is file name
class FileField(Field):
    field_type_name = FILE_TYPE
    support_param = False

    def __init__(self, file_name=None, work_path=None):
        super().__init__(file_name)
        self._work_path = work_path

    @classmethod
    def from_json(cls, json_object, input_name, field_desc: FieldDesc):
        return FileField(json_object['Value Data'], field_desc.work_path)

    def type_assert(self):
        assert isinstance(self._value, str)

    def get_input(self):
        """return absolute path"""
        return os.path.join(self._work_path, self._value).replace("\\", '/')


# _value is file path
class JsonField(Field, ListSubFieldBase):
    field_type_name = JSON_TYPE
    support_param = False

    def __init__(self, value, filename=None, work_path=None):
        super().__init__(value)
        self._filename = filename
        self._work_path = work_path

    @classmethod
    def from_json(cls, json_object, input_name, field_desc: FieldDesc):
        filename = json_object['Value Data'][0]
        file_path = os.path.join(field_desc.work_path, filename).replace("\\", '/')
        f = open(file_path, 'r')
        value = json.loads(f.read())
        return cls(value, filename, field_desc.work_path)

    def get_output(self, json_field, field_desc: FieldDesc, data_field='Value Data', output_index=0):
        if self._filename is None:
            self._filename = "jsonfile_" + json_field['label'] + ".json"
        # save file
        file_path = os.path.join(field_desc.work_path, self._filename).replace('\\', '/')
        f = open(file_path, "w")
        json.dump(self._value, f)
        # fill data field
        json_field[data_field] = [self._filename]

    @classmethod
    def list_to_json_value(cls, val, work_path=None):
        if work_path is None:
            raise Exception("json file as list sub field need work_path")
        file_name = "jsonfile_" + time.strftime("%H%M%S", time.localtime()) + str(next(unique_name_generator)) + ".json"
        file_path = os.path.join(work_path, file_name).replace("\\", '/')
        f = open(file_path, 'w')
        json.dump(val, f)
        return file_name

    @classmethod
    def list_from_json_value(cls, json_value, work_path=None):
        if work_path is None:
            return {}
        file_path = os.path.join(work_path, json_value).replace("\\", "/")
        f = open(file_path, 'r')
        value = json.loads(f.read())
        f.close()
        return value


class ListMeta(FieldMeta):
    """to support sth like 'ListField[Int]' """

    def __getitem__(self, item):
        assert issubclass(item, ListSubFieldBase)
        cls_name = 'ListField' + item.field_type_name
        cls_dict = {cls.__name__: cls for cls in ListField.__subclasses__()}
        if cls_name in cls_dict:
            return cls_dict[cls_name]
        a = ListMeta('ListField' + item.field_type_name, (ListField,), {'inner_field': item})
        a.inner_field = item
        return a


class ListField(Field, metaclass=ListMeta):
    if_list = True
    inner_field = Field
    field_type = LIST_TYPE
    support_param = False

    def __init__(self, values, inner_field_type=None):
        super().__init__(values)
        self.inner_field_type = self.inner_field if inner_field_type is None else inner_field_type

    @classmethod
    def from_json(cls, json_object, input_name, field_desc: FieldDesc):
        values = json_object["Value Data"]
        inner_field_type = FieldManager()[json_object['Data Type']]

        def helper(value):
            return inner_field_type.list_from_json_value(value, field_desc.work_path)

        values = list(map(helper, values))
        return ListField(values, inner_field_type)

    @classmethod
    def create_json_field(cls, field_category=FieldCategory.Input, name: str = None):
        field = {'Type': cls.field_type, 'Data Type': cls.inner_field.field_type_name, 'If List': cls.if_list,
                 'FieldCategory': field_category.name}
        if name is not None:
            field["Label"] = name
        return field

    def get_output(self, json_field, field_desc: FieldDesc, data_field='Value Data', output_index=0):
        def helper(value):
            return self.inner_field_type.list_to_json_value(value, work_path=field_desc.work_path)

        values = list(map(helper, self._value))
        json_field[data_field] = values



# -*- coding: utf-8 -*-
"""
Author  : NextPCG
"""

import logging
from typing import Dict, List, Union, Tuple
from .const import *
from .field import FieldManager, Field, FieldDesc
from .meta_helper import Singleton
from .dson import DsonMetaInfo, DsonManager


class Dispatcher(metaclass=Singleton):
    def __init__(self):
        pass

    @staticmethod
    def prepare_inputs(inputs_json: Dict, field_desc: FieldDesc):
        """get inputs from Input:{} json object
        """
        inputs = []
        input_index = 0
        while input_index < len(inputs_json):
            input_index += 1
            input_name = dson_input_tag + str(input_index)
            json_object = inputs_json[dson_input_tag + str(input_index)]
            inputs.append(FieldManager().create_field(json_object, input_name, field_desc))
        return inputs

    @staticmethod
    def prepare_params(params_json: Dict, field_desc: FieldDesc):
        """prepare params from json object in 'Params: {}'
        """
        params = {}
        for name, json_object in params_json.items():
            params[name] = FieldManager().create_field(json_object, name, field_desc)
        return params

    @staticmethod
    def prepare_outputs(outputs_json: Dict, results: Union[Tuple[Field], Field], field_desc: FieldDesc):
        if results is None:
            raise Exception("none results to prepare")
        if isinstance(results, tuple):
            for index in range(len(results)):
                results[index].get_output(outputs_json[dson_output_tag + str(index + 1)], field_desc)
        else:
            results.get_output(outputs_json[dson_output_tag + "1"], field_desc)

    def run_func(self, json_data: Dict, logger: logging.Logger, work_path: str, json_error_data: Dict):
        """find function in mod and run it

        side effects: change json_data
        """
        # clear error info
        if 'error' in json_data:
            del json_data['error']
        # get raw compress
        raw_compress = False
        if 'raw_compress' in json_data:
            raw_compress = bool(json_data['raw_compress'])
        # get io mode
        io_mode = 1
        if 'io_mode' in json_data:
            io_mode = int(json_data['io_mode'])

        # create field_desc
        field_desc = FieldDesc(work_path, io_mode, raw_compress)

        # find dson class
        dson_meta_info = self.get_dson_meta_info(json_data)
        cls_name = dson_meta_info.cls_name
        if cls_name not in DsonManager().dsonMap:
            raise Exception("not known class name: %s" % cls_name)
        mod = DsonManager().dsonMap[cls_name]
        func_tag = mod.label
        for json_key, json_value in json_data.items():
            if func_tag in json_key:
                inputs = []
                try:
                    logger.info("Start handing inputs...")
                    if dson_inputs_tag in json_value:
                        inputs = self.prepare_inputs(json_value[dson_inputs_tag], field_desc)
                    logger.info("Finish handing inputs!")
                except Exception as e:
                    json_error_data['Error_Tag'] = 'dispatcher handing inputs error'
                    json_error_data['Error_Info'] = str(e)
                    logger.exception(e)
                    return
                params = {}
                try:
                    logger.info("Start handing params...")
                    if dson_params_tag in json_value:
                        params = self.prepare_params(json_value[dson_params_tag], field_desc)
                    logger.info("Start handing params!")
                except Exception as e:
                    json_error_data['Error_Tag'] = 'dispatcher handing params error'
                    json_error_data['Error_Info'] = str(e)
                    logger.exception(e)
                    return
                results = None
                try:
                    # exec function
                    logger.info("Start calling functions...")
                    func_name = json_key
                    func_name = func_name.split('.')[0]
                    func = mod.__dict__[func_name].__func__
                    results = func(*inputs, **params)
                    logger.info("Finish calling functions!")
                except Exception as e:
                    json_error_data['Error_Tag'] = 'dispatcher error in running function'
                    json_error_data['Error_Info'] = str(e)
                    logger.exception("running function error")
                    return
                try:
                    # handing outputs
                    logger.info("Start handing outputs...")
                    if dson_outputs_tag in json_value:
                        self.prepare_outputs(json_value[dson_outputs_tag], results, field_desc)
                    logger.info("Finish handing outputs!")
                except Exception as e:
                    json_error_data['Error_Tag'] = 'dispatcher error in handing outputs'
                    json_error_data['Error_Info'] = str(e)
                    logger.exception("error in handing outputs")
                    return

    @staticmethod
    def get_dson_meta_info(json_data):
        return DsonMetaInfo.from_dson(json_data)


# for test
if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-
"""
Author  : NextPCG
"""

from .meta_helper import Singleton
from .const import *


class nextpcgmethod(staticmethod):
    def __init__(self, function):
        super(nextpcgmethod, self).__init__(function)


class DsonMetaInfo:
    """some wrapped meta info in dson file"""
    def __init__(self, cls_name:str, if_in_server:bool = False, dson_main_path:str = None):
        self.cls_name = cls_name
        self.if_in_server = if_in_server
        if(if_in_server == False):
            self.plugin_name = dson_main_path
        else:
            self.plugin_name = None           

    def to_json(self):
        json_data = {dson_meta_clsname_tag: self.cls_name, dson_meta_if_in_server_tag: self.if_in_server}
        if not self.if_in_server:
            json_data[dson_meta_plugin_name] = self.plugin_name
        return json_data

    @staticmethod
    def from_json(json_data):
        clsname = json_data[dson_meta_clsname_tag]
        if_in_server = True
        if dson_meta_if_in_server_tag in json_data:
            if_in_server = json_data[dson_meta_if_in_server_tag]
        dson_main_path = None
        if not if_in_server:
            dson_main_path = json_data[dson_meta_plugin_name]
        return DsonMetaInfo(clsname, if_in_server, dson_main_path)

    @staticmethod
    def from_dson(json_data):
        """from full dson file json"""
        for _, json_value in json_data.items():
            if dson_meta_field_tag in json_value:
                return DsonMetaInfo.from_json(json_value[dson_meta_field_tag])
        raise Exception("can't find dson meta info ")


class DsonManager(metaclass=Singleton):
    def __init__(self):
        self.dsonMap = {}


class DsonMeta(type):
    def __init__(cls, *args, **kwargs):
        assert hasattr(cls, 'label')
        DsonManager().dsonMap[cls.__name__] = cls
        super().__init__(*args, **kwargs)


class DsonBase(metaclass=DsonMeta):
    label = "pda"
    if_in_server = False # whether this class in NextPCG Server

    def __init__(self):
        pass



# -*- coding: utf-8 -*-
"""simple protocol for dispatch to user defined plugin
Author  : NextPCG
"""

import os
import json
import threading
import asyncio
from concurrent.futures import Future, ThreadPoolExecutor

PROTOCOL_DONE = "s8xc32ds5f" # abitray code


def module_send_done():
    print(PROTOCOL_DONE)


async def server_check_done(proc, logger):
    """check whether task is done
    
    Args:
        proc(subprocess.Popen): process running dson_main.py """
    a = await run_as_daemon(proc.stdout.readline)
    while not PROTOCOL_DONE in a:
        if not a:
            # check if proc is dead
            if proc.poll() is not None:
                # proc is terminated
                raise ProcessLookupError("subprocess is terminated by unknown bug")
        logger.info(a)
        a = await run_as_daemon(proc.stdout.readline)


def server_send_dson(dson_data, work_path, proc):
    json_data_temp_name = os.path.join(work_path, "temp.json").replace('\\','/')
    if not os.path.exists(work_path):
        os.makedirs(work_path, exist_ok=True)
    with open(json_data_temp_name, "w", encoding="utf-8") as f:
        json.dump(dson_data, f)
    proc.stdin.write(json_data_temp_name + " " + work_path + "\n")
    proc.stdin.flush()


def server_load_dson(work_path):
    dson_error_data = {}
    json_data_temp_name = os.path.join(work_path, "temp.json").replace('\\','/')
    with open(json_data_temp_name, "r", encoding='utf-8') as f:
        dson_data = json.load(f)
    json_error_data_name = os.path.join(work_path, 'error.json').replace('\\','/')
    if os.path.exists(json_error_data_name):
        with open(json_error_data_name, 'r', encoding='utf-8') as f:
            dson_error_data = json.load(f)
    return dson_data, dson_error_data


async def run_as_daemon(func, *args):
    future = Future()
    future.set_running_or_notify_cancel()

    def daemon():
        try:
            result = func(*args)
        except Exception as e:
            future.set_exception(e)
        except BaseException as e:
            future.set_exception(e)
        else:
            future.set_result(result)
    
    threading.Thread(target=daemon, daemon=True).start()
    return await asyncio.wrap_future(future)


# -*- coding: utf-8 -*-
"""
Author  : NextPCG
"""

from inspect import signature, Signature
from functools import wraps
from typing import *
from .field import Field, FieldCategory
import json
from .const import *
import logging
from .dson import DsonMetaInfo


def create_dson_from_pda(func: Callable, tag, dson_meta_info: DsonMetaInfo) -> Dict:
    cda_json = {}
    try:
        func_sig = signature(func)
        func_name = func.__name__
        inputs, params = get_inputs(func_sig)
        outputs = get_outputs(func_sig)
        doc = get_doc(func)
        cda_json_cda = {dson_inputs_tag: inputs, dson_params_tag: params,
                        dson_outputs_tag: outputs, dson_doc_tag: doc,
                       dson_meta_field_tag: dson_meta_info.to_json()}
        cda_json = {func_name + "." + tag: cda_json_cda}
    except Exception as e:
        logging.exception(e)
    finally:
        return cda_json


def get_inputs(sig: Signature):
    inputs = {}
    params = {}
    input_index = 0
    for param in sig.parameters.values():
        if param.default is param.empty:
            # no default value, we put it in inputs
            param_type = param.annotation
            assert issubclass(param_type, Field)
            field = param_type.create_json_field(FieldCategory.Input, param.name)
            input_index += 1
            inputs[dson_input_tag + str(input_index)] = field
        else:
            # have default value, we put it in params
            param_type = param.annotation
            field = param_type.create_json_field(FieldCategory.Param, param.name)
            if not param_type.support_param:
                raise Exception(str(param_type) + " can't be param")
            # param_value = param_type(param.default)
            # param_value.get_output(field, "Default Value")
            param_type.get_param_default(field, param.default)
            params[param.name] = field
    return inputs, params


def get_outputs(sig: Signature):
    outputs = {}
    return_annotation = sig.return_annotation
    output_index = 0
    if return_annotation is None:
        return outputs
    if hasattr(return_annotation, '__origin__') and (return_annotation.__origin__ == tuple or return_annotation.__origin__ == Tuple):
        for ele_type in return_annotation.__args__:
            output_index += 1
            outputs["output" + str(output_index)] = ele_type.create_json_field(FieldCategory.Output)
    else:
        output_index += 1
        outputs['output' + str(output_index)] = return_annotation.create_json_field(FieldCategory.Output)
    return outputs

def get_doc(sig: Callable):
    return sig.__doc__

# for test
if __name__ == "__main__":
    from field import Int, Int2, String, ListField


    def foo(a: Int, b: ListField[Int2] = [1, 2], c: Int2 = (1, 2)) -> String:
        pass


    jj = create_dson_from_pda(foo)
    file_path = "test.json"
    with open(file_path, "w") as fp:
        json.dump(jj, fp)



# -*- coding: utf-8 -*-
"""
Author  : NextPCG
"""
from .field import Field, FieldDesc
from .geo_texture import Texture


class TextureField(Field):
    """
    _value is a Texture
    """
    field_type_name = "paramType.Texture"
    support_param = False

    def __init__(self, value=None):
        self._value: Texture = value

    @classmethod
    def from_json(cls, json_object, input_name, field_desc: FieldDesc):
        if field_desc.work_path is None:
            raise Exception("TextureField from_json failed:work path need to be set, not allowed")
        value = Texture.create_from_json(json_object, field_desc.work_path, field_desc.io_mode, field_desc.raw_compress, input_name)
        return cls(value)

    def get_input(self):
        return self._value

    def get_output(self, json_field, field_desc: FieldDesc, data_field='Value Data', output_index=0):
        geo_index = 0
        self._value.to_json(geo_index, output_index, json_field, field_desc.io_mode, field_desc.work_path, field_desc.raw_compress)



# -*- coding: utf-8 -*-
"""metaclass for general purpose
Author  : NextPCG
"""

import weakref

'''
Singleton
'''


class Singleton(type):
    def __init__(cls, *args, **kwargs):
        cls.__instance = None
        super().__init__(*args, **kwargs)

    def __call__(cls, *args, **kwargs):
        if cls.__instance is None:
            cls.__instance = super().__call__(*args, **kwargs)
            return cls.__instance
        else:
            return cls.__instance


'''
Cached
'''


class CachedMeta(type):
    def __init__(cls, *args, **kwargs):
        super().__init__(*args, **kwargs)
        cls.__cache = weakref.WeakValueDictionary()

    def __call__(cls, *args):
        if args in cls.__cache:
            return cls.__cache[args]
        else:
            obj = super().__call__(*args)
            cls.__cache[args] = obj
            return obj



# -*- coding: utf-8 -*-
"""
Author  : NextPCG
"""

import datetime
import json
import zlib
from typing import Dict, List
import os
from PIL import Image
from .geo import GeoBase, save_output_file, write_attribs_to_file, add_custom_attributes

import numpy as np


class Volume(GeoBase):
    def __init__(self):
        super(Volume, self).__init__()
        self.name = ""
        self.volume_info = {}
        self.volume_transform = None
        self.volume_bound_center = None
        self.volume_bound_size = None
        self.mat = None

    def to_json(self, io_mode=1, work_path=None):
        json_data = super().to_json(io_mode, work_path)
        json_data['volume_info'] = self.volume_info
        # TODO: whether correct? x,y 's order
        json_data['xsize'] = np.size(self.mat, 0)
        json_data['ysize'] = np.size(self.mat, 1)
        json_data['volume_name'] = self.name
        if self.volume_bound_size is not None:
            json_data['volume_bound_size'] = self.volume_bound_size
        if self.volume_transform is not None:
            json_data['volume_transform'] = self.volume_transform
        if self.volume_bound_center is not None:
            json_data['volume_bound_center'] = self.volume_bound_center
        return json_data


class HeightField():

    def __init__(self, name=None):
        self.volumes: Dict[str, Volume] = {}

    @staticmethod
    def create_from_json(json_object: Dict, geo_key_list: List[str], work_path: str, io_mode: int, raw_compress: bool):
        heightfield = HeightField()
        geo_idx = -1
        for geo_key in geo_key_list:
            if "Geo_" in geo_key:
                volume = Volume()
                geo_val = json_object[geo_key]
                # geo_name = input_name + "_" + re.split('-|\.', geo_val)[0]
                geo_idx = geo_idx + 1
                geo_path = os.path.join(work_path, geo_val).replace("\\", "/")
                f = open(geo_path, "r", encoding="utf-8")
                geo_json_data = json.loads(f.read())["Output_Geo"]

                hf_size_x = geo_json_data['xsize']
                hf_size_y = geo_json_data['ysize']
                hf_volume_size = 1.0

                if 'volume_transform' in geo_json_data:
                    volume_transform = geo_json_data['volume_transform']

                # add volume
                hf_volume_name = geo_json_data['volume_name']
                hf_volume = np.asarray([]).astype(np.float32)
                volume_filename = os.path.join(work_path, geo_json_data['volume_file'])
                # support raw file
                packed_volume = None
                if volume_filename.endswith('.raw'):
                    dt = None
                    if 'height' in hf_volume_name or 'alphablend' in hf_volume_name:
                        dt = np.uint16
                    else:
                        dt = np.uint8
                    raw = None
                    if raw_compress:
                        with open(volume_filename, 'rb') as f_volume:
                            compress_buffer = f_volume.read()
                            raw = np.frombuffer(zlib.decompress(compress_buffer), dtype=dt)
                    else:
                        raw = np.fromfile(volume_filename, dtype=dt)
                    # reshape to actual size
                    packed_volume = np.transpose(np.reshape(np.asarray(raw), (hf_size_y, hf_size_x, 1)),[1,0,2])
                else:
                    with Image.open(volume_filename) as image:
                        packed_volume = np.transpose(np.reshape(np.asarray(image)), (hf_size_y, hf_size_x, 1), [1,0,2])

                # x y z in packed_volume

                # https://docs.unrealengine.com/4.26/en-US/BuildingWorlds/Landscape/TechnicalGuide/
                # preprocess to get real height
                zscale = 1.0
                if 'height' in hf_volume_name:
                    if 'volume_transform' in geo_json_data:
                        volume_transform = geo_json_data['volume_transform']
                        zscale = volume_transform[5]
                    hf_volume = np.vectorize(lambda x: (float(x) - 32768.0) * (zscale / 128.0))(packed_volume).astype(
                        np.float32)
                elif 'alphablend' in hf_volume_name:
                    hf_volume = np.vectorize(lambda x: (float(int(x) & 0xFFFC)) / 65532.0)(packed_volume).astype(
                        np.float32)
                else:
                    hf_volume = np.vectorize(lambda x: (x / 255.0))(packed_volume).astype(np.float32)
                volume.mat = hf_volume
                # get attribs
                # volume.detail = geo_json_data['attribs']['detail']
                # volume.vertex = geo_json_data['attribs']['vertex']
                # volume.point = geo_json_data['attribs']['point']
                # volume.prim = geo_json_data['attribs']['prim']
                volume.volume_info = geo_json_data['volume_info']
                volume.volume_transform = geo_json_data['volume_transform']
                volume.volume_bound_center = geo_json_data['volume_bound_center']
                volume.volume_bound_size = geo_json_data['volume_bound_size']
                volume.name = hf_volume_name
                add_custom_attributes(volume, geo_json_data, [], io_mode, work_path)

                heightfield.volumes[hf_volume_name] = volume
        return heightfield

    def to_json(self, geo_index: int, output_index, json_field, io_mode=1, work_path=None, raw_compress = False):
        """
        Returns:
            num of volumes
        """
        for volume_name, volume in self.volumes.items():
            json_geo = volume.to_json()
            json_geo_index = 'Geo_{0}'.format(geo_index)
            volume_filename = '{0}_Heightfield{1}_{2}_{3}.raw'.format(json_geo_index, output_index, volume_name,
                                                                      datetime.datetime.now().strftime("%y%m%d_%H%M%S"))
            json_geo['volume_file'] = volume_filename
            json_img_index = 'Img_{0}'.format(geo_index)
            json_field[json_img_index] = volume_filename
            if io_mode == 1:
                json_geo['attribs_filename'] = ''
                json_field[json_geo_index] = save_output_file(work_path,
                                                              '{0}_Heightfield{1}'.format(json_geo_index, output_index),
                                                              json_geo)
            elif io_mode == 2:
                attri_geo_index = 'Attr_{0}'.format(geo_index)
                attribs_filename = '{0}_Heightfield{1}_{2}.attr_raw'.format(json_geo_index, output_index,
                                                                            datetime.datetime.now().strftime(
                                                                                "%y%m%d_%H%M%S"))
                json_geo['attribs_filename'] = attribs_filename
                attribs_filepath = os.path.join(work_path, attribs_filename)
                write_attribs_to_file(attribs_filepath, json_geo)
                json_field[json_geo_index] = save_output_file(work_path,
                                                              '{0}_HeightField{1}'.format(json_geo_index, output_index),
                                                              json_geo)
                json_field[attri_geo_index] = attribs_filename
            else:
                raise Exception("unsupported io mode")

            save_path = os.path.join(work_path, volume_filename)

            packed_volume = np.transpose(volume.mat)
            if 'height' in volume_name:
                volume_zscale = json_geo['volume_info']['transform']['scale'][2]
                packed_volume = np.vectorize(lambda x: np.uint16(x / volume_zscale * 128.0 + 32768.0))(packed_volume)
                image = Image.fromarray(packed_volume.astype(np.uint16), 'I;16')
                image.save(save_path.replace('.raw', '.tiff'))
                if raw_compress:
                    with open(save_path, 'ab+') as f_volume:
                        buffer = packed_volume.tobytes()
                        compress_buffer = zlib.compress(buffer)
                        f_volume.write(compress_buffer)
                else:
                    packed_volume.tofile(save_path)
            elif 'alphablend' in volume_name:
                pass
                # kivlin. UE4.27 not support to set height alpha blend map.
            else:  # weight
                pass
                packed_volume = np.vectorize(lambda x: np.uint8(x * 255.0))(packed_volume)
                # image = Image.fromarray(packed_volume.astype(np.uint8))
                # image.save(save_path)
                if raw_compress:
                    with open(save_path, 'ab+') as f_volume:
                        buffer = packed_volume.tobytes()
                        compress_buffer = zlib.compress(buffer)
                        f_volume.write(compress_buffer)
                else:
                    # image = Image.fromarray(packed_volume.astype(np.uint8))
                    # image.save(save_path)
                    packed_volume.tofile(save_path)
            geo_index += 1
        return len(self.volumes)


# -*- coding: utf-8 -*-
"""
Author  : NextPCG
"""

from .field import Field, FieldCategory, FieldDesc
from .geo_instanced_staticmesh import InstancedStaticMesh


class InstancedStaticMeshField(Field):
    """
    _value is a InstanceStaticMesh
    """
    field_type_name = "paramType.InstancedStaticMesh"
    support_param = False

    def __init__(self, value):
        self._value = value

    @classmethod
    def from_json(cls, json_object, input_name, field_desc:FieldDesc):
        if field_desc.work_path is None:
            raise Exception("HeightFieldField from_json failed:work path need to be set, not allowed")
        value = InstancedStaticMesh.create_from_json(json_object, field_desc.work_path, field_desc.io_mode, input_name)
        return InstancedStaticMeshField(value)

    def get_input(self):
        return self._value

    def get_output(self, json_field, field_desc, data_field='Value Data', output_index=0):
        geo_index = 0
        self._value.to_json(geo_index, output_index, json_field, field_desc.io_mode, field_desc.work_path)



# -*- coding: utf-8 -*-
"""
Author  : NextPCG
"""

import datetime
import json
import re
from typing import Dict, List
import os
from PIL import Image
from .geo import GeoBase, save_output_file, write_attribs_to_file, load_attribs_from_file, add_custom_attributes

import numpy as np


class InstanceNode(GeoBase):
    """simulate Instance node on houdini"""

    def __init__(self):
        super().__init__()
        self.name: str = ""
        self.positions = []  #
        self.orientations = []  # : quaternion
        self.scales = []
        self.instanced_geos = None

    def to_json(self, io_mode=1, work_path=None):
        json_data = super().to_json(io_mode, work_path)

        # don't support mesh geo now TODO
        # if self.instanced_geos is not None:
        #     json_data['instance_geos'] = self.instanced_geos

        json_mesh_point_data = json_data['attribs']['point']

        if len(self.positions) > 0:
            json_mesh_point_data['P'] = {'type': 1, 'data': self.positions}
        if len(self.orientations) > 0:
            json_mesh_point_data['orient'] = {'type': 1, 'data': self.orientations}
        if len(self.scales) > 0:
            json_mesh_point_data['scale'] = {'type': 1, 'data': self.scales}
        # transform

        if io_mode == 1 or io_mode == 0:
            json_trans = {}
            for point_index in range(len(self.positions)):
                xform = {}
                xform['position'] = json_mesh_point_data['P']['data'][point_index]
                # support default orient and scale for instance.
                orient = [0, 0, 0, 1]
                if 'orient' in json_mesh_point_data:
                    orient = json_mesh_point_data['orient']['data'][point_index]
                scale = [1, 1, 1]
                if 'scale' in json_mesh_point_data:
                    scale = json_mesh_point_data['scale']['data'][point_index]
                xform['rotationQuaternion'] = orient
                xform['scale'] = scale
                xform['sheer'] = [0, 0, 0]
                xform['rstorder'] = 5
                json_trans[str(point_index)] = xform
            json_data['transforms'] = json_trans

        return json_data


def create_instance_node(geo_json_data, geo_name, io_mode, work_path) -> InstanceNode:
    instance_node = InstanceNode()

    new_vertices = []
    new_points = []
    new_prims = []

    # create instance.
    pos_buffer = []
    orient_buffer = []
    scale_buffer = []
    if "attribs" in geo_json_data:
        input_val = geo_json_data['attribs']
        if 'P' in input_val['point'] and 'orient' in input_val['point'] and 'scale' in input_val['point']:
            if io_mode == 0 or io_mode == 1:
                pos_buffer = (input_val['point']['P']['data'])
                orient_buffer = (input_val['point']['orient']['data'])
                scale_buffer = (input_val['point']['scale']['data'])
            elif io_mode == 2 or io_mode == 3:
                pos_buffer = load_attribs_from_file(geo_json_data['attribs_filename'], input_val['point']['P'],
                                                    work_path)
                orient_buffer = load_attribs_from_file(geo_json_data['attribs_filename'], input_val['point']['orient'],
                                                       work_path)
                scale_buffer = load_attribs_from_file(geo_json_data['attribs_filename'], input_val['point']['scale'],
                                                      work_path)

        # not support staticMesh geo yes
        if 'instanced_geos' in geo_json_data:
            instance_node.instanced_geos = geo_json_data['instanced_geos']

    instance_node.positions = pos_buffer
    instance_node.orientations = orient_buffer
    instance_node.scales = scale_buffer

    add_custom_attributes(instance_node, geo_json_data, ['P'], io_mode, work_path)
    return instance_node


class InstancedStaticMesh:
    def __init__(self):
        self.instance_nodes: Dict[str, InstanceNode] = {}

    @staticmethod
    def create_from_json(json_field: Dict, work_path: str, io_mode: int, input_name: str):
        instance_staticmesh = InstancedStaticMesh()
        for geo_key, geo_val in json_field.items():
            if 'Geo_' in geo_key:
                geo_name = input_name + '_' + re.split('-|\.', geo_val)[0]
                geo_path = os.path.join(work_path, geo_val)
                f = open(geo_path, 'r', encoding='utf-8')
                geo_json_data = json.loads(f.read())['Output_Geo']
                f.close()

                # create instance node
                instance_staticmesh.instance_nodes[geo_name] = create_instance_node(geo_json_data, geo_name, io_mode,
                                                                                    work_path)

        return instance_staticmesh

    def to_json(self, geo_index: int, output_index, json_field, io_mode=1, work_path=None):
        for geo_name, instance_node in self.instance_nodes.items():
            json_geo = instance_node.to_json()
            json_geo_index = 'Geo_{0}'.format(geo_index)
            if io_mode == 1 or io_mode == 0:
                json_geo['attribs_filename'] = ''
                json_field[json_geo_index] = save_output_file(work_path,
                                                              '{0}_Instance{1}'.format(json_geo_index, output_index),
                                                              json_geo)
            elif io_mode ==2 or io_mode==3:
                attri_geo_index = "Attr_{0}".format(geo_index)
                attribs_filename = '{0}_Instance{1}_{2}.attr_raw'.format(json_geo_index, output_index, datetime.datetime.now().strftime("%y%m%d_%H%M%S"))
                json_geo['attribs_filename'] = attribs_filename
                abs_attribs_filename = os.path.join(work_path, attribs_filename)
                offset = write_attribs_to_file(abs_attribs_filename, json_geo)

                # mesh geo todo

                json_field[json_geo_index] = save_output_file(work_path,
                                                              '{0}_Instance{1}'.format(json_geo_index, output_index),
                                                              json_geo)
                json_field[attri_geo_index] = attribs_filename
            else:
                raise Exception("unsupported io mode")

            geo_index += 1
        return len(self.instance_nodes)



# -*- coding: utf-8 -*-
"""
Author  : NextPCG
"""

import datetime
import json
import os.path
import re
import zlib
import numpy as np

from typing import Dict, List
from PIL import Image
from .geo import GeoBase, add_custom_attributes, save_output_file, write_attribs_to_file


class TextureNode(GeoBase):

    def __init__(self):
        super(TextureNode, self).__init__()
        self.name: str = ''
        self.volume_info = dict()
        self.bit_depth = 'u8'
        self.channel_count = 0
        self.mat = {
            'r': None,
            'g': None,
            'b': None,
            'a': None,
        }

    def to_json(self, io_mode=1, work_path=None):
        # get each level attributes
        json_data = super().to_json(io_mode, work_path)
        json_data['volume_info'] = self.volume_info
        json_data['volume_name'] = self.name
        json_data['xsize'] = np.size(self.mat['r'], 1)
        json_data['ysize'] = np.size(self.mat['r'], 0)
        json_data['zsize'] = 1
        json_data['bit_depth'] = self.bit_depth
        # channel = 0
        # for mat in self.mat:
        #     if mat:
        #         channel += 1
        json_data['channel_count'] = self.channel_count
        return json_data


def create_texture_node(geo_json_data, work_path, io_mode, raw_compress):
    # volume info
    hf_size_x = geo_json_data['xsize']
    hf_size_y = geo_json_data['ysize']
    hf_volume_name = geo_json_data['volume_name']
    volume_filename = os.path.join(work_path, geo_json_data['volume_file'])

    # bit depth
    bit_depth_remap = dict()
    bit_depth_remap['u8'] = (np.uint8, 255)
    bit_depth_remap['u16'] = (np.uint16, 65535)
    bit_depth_remap['u32'] = (np.uint16, 4294967295)
    bit_depth_remap['f16'] = (np.uint16, 1.0)
    bit_depth_remap['f32'] = (np.uint32, 1.0)

    bit_depth = None
    bit_type = None
    bit_scale = 0
    channel_count = 0

    # check is raw
    is_raw = False
    base_name = os.path.basename(volume_filename)

    if base_name.endswith('.raw'):
        is_raw = True
        bit_depth = geo_json_data['bit_depth']
        channel_count = geo_json_data['channel_count']
        bit_type, bit_scale = bit_depth_remap[bit_depth]

    # support raw file
    raw = None
    if is_raw:
        if raw_compress:
            with open(volume_filename, 'rb') as f_volume:
                compress_buffer = f_volume.read()
                raw = np.asarray(np.frombuffer(zlib.decompress(compress_buffer), dtype=bit_type))
        else:
            raw = np.asarray(np.fromfile(volume_filename, dtype=bit_type))
    else:
        with Image.open(volume_filename) as image:
            channel_count = len(image.getbands())
            hf_size_x, hf_size_y = image.size
            raw = np.asarray(image)
            bit_type = raw.dtype
            if np.issubdtype(bit_type, np.integer):
                bit_scale = np.iinfo(bit_type).max
            else:
                bit_scale = 1.0
            for bit_depth_key in bit_depth_remap.keys():
                if bit_type.type is bit_depth_remap[bit_depth_key][0]:
                    bit_depth = bit_depth_key
                    break

    '''
        the transmission in server_houdini is as follow:
            1. ExportTexture(in UE) -> this step flip the y axis for raw data 
            2. Houdini Function
            3. ImportTexture(in UE) -> this step re-flip the y axis for raw data

    '''
    # cause raw data has been flipped in ExportTexture(in UE),so we flipped back here
    if is_raw:
        raw = np.flip(np.reshape(raw, (hf_size_y, hf_size_x, channel_count)), 0)

    packed_volume = np.reshape(raw, (hf_size_x * hf_size_y, channel_count)).astype(np.float32) / bit_scale
    # split channels
    split_volumes = np.hsplit(packed_volume, channel_count)

    # create texture node
    texture_node = TextureNode()
    texture_node.name = hf_volume_name
    texture_node.volume_info = geo_json_data['volume_info']
    texture_node.bit_depth = bit_depth
    texture_node.channel_count = len(split_volumes)
    for i in range(len(split_volumes)):
        if i == 0:
            texture_node.mat['r'] = np.reshape(split_volumes[i], (hf_size_y, hf_size_x))
        elif i == 1:
            texture_node.mat['g'] = np.reshape(split_volumes[i], (hf_size_y, hf_size_x))
        elif i == 2:
            texture_node.mat['b'] = np.reshape(split_volumes[i], (hf_size_y, hf_size_x))
        elif i == 3:
            texture_node.mat['a'] = np.reshape(split_volumes[i], (hf_size_y, hf_size_x))
    add_custom_attributes(texture_node, geo_json_data, [], io_mode, work_path)

    return texture_node


def image2node(image:Image.Image, name:str):

    # bit depth
    bit_depth_remap = dict()
    bit_depth_remap['u8'] = (np.uint8, 255)
    bit_depth_remap['u16'] = (np.uint16, 65535)
    bit_depth_remap['u32'] = (np.uint16, 4294967295)
    bit_depth_remap['f16'] = (np.uint16, 1.0)
    bit_depth_remap['f32'] = (np.uint32, 1.0)


    channel_count = len(image.getbands())
    size_x, size_y = image.size
    raw = np.asarray(image)
    bit_type = raw.dtype
    if np.issubdtype(bit_type, np.integer):
        bit_scale = np.iinfo(bit_type).max
    else:
        bit_scale = 1.0
    for bit_depth_key in bit_depth_remap.keys():
        if bit_type.type is bit_depth_remap[bit_depth_key][0]:
            bit_depth = bit_depth_key
            break

    packed_volume = np.reshape(raw, (size_x * size_y, channel_count)).astype(np.float32) / bit_scale
    # split channel
    split_volumes = np.hsplit(packed_volume, channel_count)

    # for i in range(len(split_volumes)):
    #     split_volumes[i] = np.flip(split_volumes[i].reshape((size_x, size_y)), 1)

    texture_node = TextureNode()
    texture_node.name = name
    texture_node.bit_depth = bit_depth
    texture_node.channel_count = 4
    # ensure alpha is add
    if len(split_volumes) == 3:
        texture_node.mat['a'] = np.ones((size_y, size_x), dtype=np.float32)
    assert len(split_volumes) >= 3, "assume image has at least 3 channels"
    for i in range(len(split_volumes)):
        if i == 0:
            texture_node.mat['r'] = np.reshape(split_volumes[i], (size_y, size_x))
        elif i == 1:
            texture_node.mat['g'] = np.reshape(split_volumes[i], (size_y, size_x))
        elif i == 2:
            texture_node.mat['b'] = np.reshape(split_volumes[i], (size_y, size_x))
        elif i == 3:
            texture_node.mat['a'] = np.reshape(split_volumes[i], (size_y, size_x))
    return texture_node


class Texture:
    def __init__(self):
        self.texture_nodes: Dict[str, TextureNode] = {}

    @staticmethod
    def create_from_json(json_field: Dict, work_path, io_mode, raw_compress, input_name: str):
        texture = Texture()
        for geo_key, geo_val in json_field.items():
            if 'Geo_' in geo_key:
                geo_name = input_name + '_' + re.split('-|\.', geo_val)[0]
                geo_path = os.path.join(work_path, geo_val)
                geo_json_data = None
                with open(geo_path, 'r', encoding='utf-8') as f:
                    geo_json_data = json.loads(f.read())['Output_Geo']
                # create texture node
                texture.texture_nodes[geo_name] = create_texture_node(geo_json_data, work_path, io_mode, raw_compress)

        return texture

    @staticmethod
    def create_from_images(images):
        texture = Texture()
        i = 1
        for image in images:
            texture.texture_nodes[f'image{i}'] = image2node(image, f'image{i}')
            i += 1
        return texture

    def to_json(self, geo_index, output_index, json_field, io_mode=1, work_path=None, raw_compress=False):
        for geo_name, texture_node in self.texture_nodes.items():
            # get base json
            json_geo = texture_node.to_json()

            # get volume filename
            json_geo_index = 'Geo_{0}'.format(geo_index)
            file_name = texture_node.name
            volume_filename = '{0}_volume{1}_{2}_{3}.raw'.format(json_geo_index, output_index, file_name,
                                                                datetime.datetime.now().strftime("%y%m%d_%H%M%S"))
            json_geo['volume_file'] = volume_filename
            json_img_index = 'Img_{0}'.format(geo_index)
            json_field[json_img_index] = volume_filename
            if io_mode == 1:
                json_geo['attribs_filename'] = ''
                json_field[json_geo_index] = save_output_file(work_path,
                                                              '{0}_Volume{1}'.format(json_geo_index, output_index),
                                                              json_geo)
            elif io_mode == 2:
                attr_geo_index = 'Attr_{0}'.format(geo_index)
                attribs_filename = '{0}_Volume{1}_{2}.attr_raw'.format(json_geo_index,
                                                                       output_index,
                                                                       datetime.datetime.now().strftime("%y%m%d_%H%M%S"))
                json_geo['attribs_filename'] = attribs_filename
                attribs_filepath = os.path.join(work_path, attribs_filename)
                write_attribs_to_file(attribs_filepath, json_geo)
                json_field[json_geo_index] = save_output_file(work_path,
                                                              '{0}_Volume{1}'.format(json_geo_index, output_index),
                                                              json_geo)
                json_field[attr_geo_index] = attribs_filename
            else:
                raise Exception("unsupported io mode")

            # bit depth map
            bit_depth_remap = {}
            bit_depth_remap['u8'] = (np.uint8, 255, True)
            bit_depth_remap['u16'] = (np.uint16, 65535, True)
            bit_depth_remap['u32'] = (np.uint16, 4294967295, True)
            bit_depth_remap['f16'] = (np.float16, 1.0, False)
            bit_depth_remap['f32'] = (np.float32, 1.0, False)

            bit_depth = texture_node.bit_depth
            numpy_type, bit_scale, is_integer = bit_depth_remap[bit_depth]

            volume_dict = dict()
            xsize = json_geo['xsize']
            ysize = json_geo['ysize']

            '''
                the transmission in server_houdini is as follow:
                    1. ExportTexture(in UE) -> this step flip the y axis for raw data 
                    2. Houdini Function
                    3. ImportTexture(in UE) -> this step re-flip the y axis for raw data
            '''
            # cause raw data should be flipped to fit transmission based on server_houdini and we choose raw-data to transmit back, so we flipped back here
            for volume_key, volume in texture_node.mat.items():
                if volume is not None:
                    volume_dict[volume_key] = np.flip(np.reshape(np.asarray(volume), (ysize, xsize)), 0)

            # lambda function to decide whether to round
            round_wrapper = None
            if is_integer:
                round_wrapper = lambda x: round(x)
            else:
                round_wrapper = lambda x: x
            combined_volume = None
            channel_count = texture_node.channel_count
            if channel_count == 1:
                r = np.vectorize(lambda x: numpy_type(round_wrapper(x * bit_scale)))(volume_dict['r'])

                combined_volume = r
            elif channel_count == 2:
                r = np.vectorize(lambda x: numpy_type(round_wrapper(x * bit_scale)))(volume_dict['r'])
                g = np.vectorize(lambda x: numpy_type(round_wrapper(x * bit_scale)))(volume_dict['g'])

                combined_volume = np.dstack((r, g))
            elif channel_count == 3:
                r = np.vectorize(lambda x: numpy_type(round_wrapper(x * bit_scale)))(volume_dict['r'])
                g = np.vectorize(lambda x: numpy_type(round_wrapper(x * bit_scale)))(volume_dict['g'])
                b = np.vectorize(lambda x: numpy_type(round_wrapper(x * bit_scale)))(volume_dict['b'])

                combined_volume = np.dstack((r, g, b))
            elif channel_count == 4:
                r = np.vectorize(lambda x: numpy_type(round_wrapper(x * bit_scale)))(volume_dict['r'])
                g = np.vectorize(lambda x: numpy_type(round_wrapper(x * bit_scale)))(volume_dict['g'])
                b = np.vectorize(lambda x: numpy_type(round_wrapper(x * bit_scale)))(volume_dict['b'])
                a = np.vectorize(lambda x: numpy_type(round_wrapper(x * bit_scale)))(volume_dict['a'])

                combined_volume = np.dstack((r, g, b, a))

            save_path = os.path.join(work_path, volume_filename)
            if raw_compress:
                with open(save_path, 'ab+') as f_volume:
                    buffer = combined_volume.tobytes()
                    compress_buffer = zlib.compress(buffer)
                    f_volume.write(compress_buffer)
            else:
                combined_volume.tofile(save_path)

            geo_index += 1



import json
import yaml
import os, sys
from pypapi import *

sys.path.append(os.path.dirname(os.path.dirname(__file__)))
base_path = os.path.dirname(__file__)

# read dson_config.yaml
dson_config = {}
with open(os.path.join(base_path, 'dson_config.yaml'), 'r', encoding='utf-8') as f:
    dson_config = yaml.safe_load(f.read())
dson_plugin_name = dson_config['name']

# get all dson class
for module in dson_config['modules']:
    if isinstance(module, str):
        # only module name
        try:
            __import__(module, globals(), locals(), [], 0)
        except Exception as e:
            print(e)
            exit(1)
    else:
        for module_name in module:
            # if has class name
            try:
                __import__(module_name, globals(), locals(), module[module_name], 0)
            except Exception as e:
                print(e)
                exit(1)


def generate_dson_files(plugin_name, dson_path):
    index = 0
    if not os.path.exists(dson_path):
        os.makedirs(dson_path, exist_ok=True)
    for cls_name, cls in DsonManager().dsonMap.items():
        for name, value in vars(cls).items():
            if isinstance(value, nextpcgmethod):
                func = value.__func__
                dson_meta_info = DsonMetaInfo(cls_name, cls.if_in_server, plugin_name)
                json_data = create_dson_from_pda(func, cls.label, dson_meta_info)
                file_name = name + '.dson'
                file_path = os.path.join(dson_path, file_name).replace('\\', '/')
                fp = open(file_path, "w")
                json.dump(json_data, fp)
                index += 1


def generate_dson_init_script(nextpcg_path, conda_env):
    if os.name != 'posix':
        # windows
        script_name = 'dson_init.bat'
        script_string = f"""@echo off
cd %~dp0 
cd ..
call conda activate {conda_env}
@REM call conda env list
python dson_main.py
"""
    else:
        # linux
        script_name = 'dson_init.sh'
        script_string = f"""#! /usr/bin/bash
source ~/.bash_profile # get conda into env
BASEDIR=$(dirname $(readlink -f "$0"))
cd "$BASEDIR"
cd ..
eval "$(conda shell.bash hook)"
conda activate {conda_env}
python dson_main.py
"""
    if not os.path.exists(nextpcg_path):
        os.makedirs(nextpcg_path, exist_ok=True)
    with open(os.path.join(nextpcg_path, script_name), 'w', encoding='utf-8') as f:
        f.write(script_string)


def generate_dson_main(base_path, dson_config):
    dsonmain_string1 ="""from pypapi import Dispatcher
from pypapi.plugin_protocol import module_send_done
import sys
import os
import json
import logging

"""
    # dsonmain_import contains code used to load necessary modules or classes
    dsonmain_import =''
    for module in dson_config['modules']:
        if isinstance(module, str):
            # only module name
            try:
                dsonmain_import+='import ' + module + '\n'
            except Exception as e:
                print(e)
                exit(1)
        else:
            for module_name in module:
                # if has class name
                dsonmain_import += 'from ' + module_name + ' import  '
                for class_name in module[module_name]:
                    dsonmain_import += class_name +','
                dsonmain_import = dsonmain_import[:-1]
    dsonmain_string2 = f"""

dson_plugin_name = '{dson_plugin_name}'
dispatcher = Dispatcher()
logger = logging.getLogger('pypapi_'+dson_plugin_name)
current_path = os.path.split(os.path.realpath(__file__))[0] # path to dosn_path.py
dot_nextpcg_path = os.path.join(current_path, '.nextpcg')

def set_up_logger():
    logFormatter = logging.Formatter('%(levelname)s:%(name)s:[%(asctime)s]: %(message)s')

    if not os.path.exists(dot_nextpcg_path):
        os.makedirs(dot_nextpcg_path)

    logfile_path = os.path.join(current_path, '.nextpcg', 'nextpcgpython.log')
    # delete old log file
    try:
        os.remove(logfile_path)
    except OSError:
        pass

    fileHandler = logging.FileHandler(logfile_path)
    fileHandler.setFormatter(logFormatter)

    consoleHandler = logging.StreamHandler()
    consoleHandler.setFormatter(logFormatter)

    logger.addHandler(fileHandler)
    logger.addHandler(consoleHandler)
    
class Tee:
    def __init__(self, *files):
        self.files = files
    
    def write(self, obj):
        for f in self.files:
            f.write(obj)
            f.flush()
    
    def flush(self):
        for f in self.files:
            f.flush()

if __name__ == '__main__':
    set_up_logger()
    ferr = open(os.path.join(dot_nextpcg_path, 'err.txt'), 'w')
    sys.stderr = Tee(sys.stderr,ferr)
    module_send_done()
    while(True):
        json_file_name, work_path = input().split()
        json_error_data = {{}}
        try:
            with open(json_file_name, 'r') as f:
                json_data = json.load(f)
            if not json_data:
                print("didn't get dson data")
                module_send_done()
            dispatcher.run_func(json_data, logger, work_path, json_error_data)
            json_data_error_file_name = os.path.join(work_path, 'error.json').replace("\\\\",'/')
            # write back json_data
            with open(json_file_name, 'w') as f:
                json.dump(json_data, f)
        except Exception as e:
            json_error_data['Error_Tag'] = 'plugin {dson_plugin_name} error in main loop'
            json_error_data['Error_Info'] = str(e)
            logger.exception(e)
        finally:
            with open(json_data_error_file_name, 'w') as f:
                json_error_data = json.dump(json_error_data, f)
        module_send_done()
"""
    dsonmain_string = dsonmain_string1 + dsonmain_import + dsonmain_string2
    with open(os.path.join(base_path, 'dson_main.py'), 'w', encoding='utf-8') as f:
        f.write(dsonmain_string)



if __name__ == "__main__":
    try:
        dson_path = os.path.join(base_path, "dson")
        nextpcg_path = os.path.join(base_path, '.nextpcg')
        generate_dson_files(dson_plugin_name, dson_path)
        generate_dson_init_script(nextpcg_path, dson_config['conda_env'])
        generate_dson_main(base_path, dson_config)        
    except Exception as e:
        print(e)
        exit(1) 
    # todo:  generate dson_main, dson_init.bat, dson_init.sh
    


name = "pypapi"

from .field import Bool, Int, Int2, Int3, Int4, Float, Float2, Float3, Float4, String, JsonField, ListField, \
    FileField, JsonField
from .field_heightfield import HeightFieldField
from .field_instanced_staticmesh import InstancedStaticMeshField
from .field_texture import TextureField

from .geo import GeoBase, save_output_file, write_attribs_to_file, add_custom_attributes
from .geo_heightfield import HeightField, Volume
from .geo_instanced_staticmesh import InstancedStaticMesh, InstanceNode

from .macro import get_fun_name

from .dispatch import Dispatcher

from .dson import DsonManager, DsonBase, nextpcgmethod, DsonMetaInfo
from .dson_create import create_dson_from_pda
from . import plugin_protocol



# -*- coding: utf-8 -*-
"""command line tool
Author  : NextPCG
"""
import argparse
import os
import yaml
import shutil
from importlib_resources import files, as_file

def entry():
    parser = argparse.ArgumentParser(prog = 'nextpcg')
    subparsers = parser.add_subparsers(
        title='These are common NextPCG Dson commands used in various situations',
        metavar='command'
    )

    init_parser = subparsers.add_parser(
        'init',
        help='Init a nextpcg python plugin environment'
    )
    init_parser.add_argument(
        'path',
        help='the path to plugin directory(optional, default is current directory)',
        nargs='?',
        default='.'
    )
    init_parser.add_argument('--name', help='the plugin name')
    init_parser.add_argument('--conda_env', help='conda env name')
    init_parser.set_defaults(handle=handle_init)

    # run parser
    args = parser.parse_args()
    if hasattr(args, 'handle'):
        args.handle(args)
    else:
        parser.print_help()


def handle_init(args):
    try:
        plugin_path = os.path.abspath(args.path)
    except Exception as e:
        print('path is no valid')
        print(e)

    dson_text = files('pypapi').joinpath('pantry').joinpath('dson_config.yaml').read_text()
    dson_config = yaml.full_load(dson_text)
    # deal with optional arg(name and conda_env)
    if hasattr(args, 'name'):
        dson_config['name'] = args.name
    if hasattr(args, 'conda_env'):
        dson_config['conda_env'] = args.conda_env
    
    with open(os.path.join(plugin_path, 'dson_config.yaml'), 'w', encoding='utf-8') as f:
        yaml.safe_dump(dson_config, f, sort_keys=False)
        f.close()
    
    source = files('pypapi').joinpath('pantry').joinpath('dson_generator.py')
    with as_file(source) as source_path:
        shutil.copyfile(source_path,os.path.join(plugin_path, 'dson_generator.py'))

    


if __name__ == '__main__':
    entry()